import datetime
import logging
import re
import sys

import pysam

logger = logging.getLogger(__name__)


class PositionsQuery(object):
    """
    Read bam input file sorted by coordinates. Read the records of one chromosom.

    Args:
        bam_input                               (str): path to input file (bam format sorted by coordinates). Must to be bam.bai file in the same folder
        output_file                             (str): path to output file
        filtered_cell_barcodes                  (dict): keys are valid cells, filtered by Seurat. values are None
        min_mapq                                (int): Minimum quality of the read mapping
        max_gene_length                         (int): Maximum length of the gene. Reads that will be mapped to longer bases will be discarded
        min_cells                               (int): mininum cells in genome position that contains the number of umis and reads according to the other parameters
        max_cells                               (int): maximum cells in genome position that contains the number of umis and reads according to the other parameters
        min_mutated_umis                        (int): mininum umis per cell that all reads contain mutation in the position
        max_mutated_umis                        (int): maximum umis per cell that all reads contain mutation in the position
        min_reads_per_mutated_umi               (int): mininum reads in at least one of umis in the cell in genome position
        max_reads_per_mutated_umi               (int): maximum reads in at least one of umis in the cell in genome position
        min_non_mutated_umis                    (int): mininum umis per cell that all reads not contain mutation in the genome position
        max_non_mutated_umis                    (int): maximum umis per cell that all reads not contain mutation in the genome position
        min_reads_per_non_mutated_umi           (int): mininum reads in at least one of umis in the cell in genome position
        max_reads_per_non_mutated_umi           (int): maximum reads in at least one of umis in the cell in genome position
        enable_cells_with_invalid_umis_num      (int): enable positions that contain cell/s with not valid umis number (according to ther range in the other parameters)
        enable_umis_with_invalid_reads_num      (int): enable positions that contain umi/s with not valid reads number (according to ther range in the other parameters)

    Attributes:
        prev_sam_record             (pysam.AlignedSegment object): save the previous read

    """

    def __init__(self, bam_input, output_file, filtered_cell_barcodes, min_mapq, max_gene_length, min_cells, max_cells,
                 min_mutated_umis, max_mutated_umis, min_reads_per_mutated_umi,
                 max_reads_per_mutated_umi, min_non_mutated_umis, max_non_mutated_umis, min_reads_per_non_mutated_umi,
                 max_reads_per_non_mutated_umi, enable_cells_with_invalid_umis_num,
                 enable_umis_with_invalid_reads_num, tag_of_umi, tag_of_cell_barcode, umi_start, umi_length,
                 cell_barcode_start, cell_barcode_length):
        self.bam_input = bam_input
        self.output_file = output_file
        self.filtered_cell_barcodes = filtered_cell_barcodes
        self.min_mapq = min_mapq
        self.max_gene_length = max_gene_length
        self.min_cells = min_cells
        self.max_cells = max_cells
        self.min_mutated_umis = min_mutated_umis
        self.max_mutated_umis = max_mutated_umis
        self.min_reads_per_mutated_umi = min_reads_per_mutated_umi
        self.max_reads_per_mutated_umi = max_reads_per_mutated_umi
        self.min_non_mutated_umis = min_non_mutated_umis
        self.max_non_mutated_umis = max_non_mutated_umis
        self.min_reads_per_non_mutated_umi = min_reads_per_non_mutated_umi
        self.max_reads_per_non_mutated_umi = max_reads_per_non_mutated_umi
        self.enable_cells_with_invalid_umis_num = enable_cells_with_invalid_umis_num
        self.enable_umis_with_invalid_reads_num = enable_umis_with_invalid_reads_num
        self.tag_of_umi = tag_of_umi
        self.tag_of_cell_barcode = tag_of_cell_barcode
        self.umi_start = umi_start
        self.umi_length = umi_length
        self.cell_barcode_start = cell_barcode_start
        self.cell_barcode_length = cell_barcode_length

        if self.bam_input:  # for the tests
            self.bam_reader = pysam.AlignmentFile(self.bam_input, "rb")
        self.prev_sam_record = None
        self.valid_positions = []

    def close_file(self):
        self.bam_reader.close()

    @staticmethod
    def get_chrom_list(bam_input):
        """
        Get the list of chromosoms from the header lines of the bam file
        Also check that the file is sorted by coordinates

        Args:
            bam_input    (str): path of the input file

        Returns:
            chrom_list    (list of str): list of chromosoms
        """

        bam_reader = pysam.AlignmentFile(bam_input, "rb")
        if bam_reader.header['HD']['SO'] != "coordinate":
            raise ValueError("Bam is not sorted by coordinate")
        chrom_list = []
        chroms_starts = []
        chroms_len = []

        try:
            for line in bam_reader.header['SQ']:
                chrom_list.append(line['SN'])
                chroms_len.append(int(line['LN']))
                chroms_starts.append(0)
        except Exception:
            raise IOError(
                "You cannot run on all genome because the bam file don't contain header lines with information on the chromosomes. You need supply coordinates manually.")
        return (chrom_list, chroms_starts, chroms_len)

    def is_mutated_read(self, pileupread):
        ref_base = pileupread.alignment.get_reference_sequence()[pileupread.query_position]
        if pileupread.alignment.query_sequence[pileupread.query_position] != ref_base:
            return ref_base
        else:
            return False

    def find_base_coverage(self, chrom='all', start=0, end=None):
        str_end = '-' + str(end) if end else ''
        print "%s Start chromomosome: %s:%s%s " % (datetime.datetime.now(), chrom, start, str_end)
        output_file_name = self.output_file + '_' + str(chrom) + '.txt' if not end else self.output_file + '_' + str(
            chrom) + '_' + str(start) + '_' + str(end) + '.txt'
        with open(output_file_name, 'w') as out_fh:
            # 6248811, 161038034
            # for pos in xrange(6248811, 6248812):
            # for pos in xrange(161038033, 161038034):
            # for pos in xrange(chrom_len):

            for pileupcolumn in self.bam_reader.pileup(contig=chrom, start=start, end=end, flag_filter=0,
                                                       ignore_orphans=False, min_mapping_quality=10,
                                                       steeper='nofilter', max_depth=100000000,
                                                       truncate=True, mark_matches=True, mark_ends=True,
                                                       add_indels=True):
                if pileupcolumn.reference_pos % 10000000 == 0:
                    print "%s: %s:%s%s %s" % (
                        datetime.datetime.now(), chrom, start, str_end, pileupcolumn.reference_pos)

                    sys.stdout.flush()

                # Filter out position with low coverage (get_num_aligned function return the reads that count by pileup
                # function except the filtered reads, but including our filtered reads).
                # if pileupcolumn.get_num_aligned() < min_total_read:
                #     continue
                pos_data = {}
                for pileupread in pileupcolumn.pileups:
                    cell_barcode = pileupread.alignment.get_tag(self.tag_of_cell_barcode)[
                                   self.cell_barcode_start:self.cell_barcode_length]
                    umi = pileupread.alignment.get_tag(self.tag_of_umi)[self.umi_start:self.umi_length]

                    if pileupread.is_del or pileupread.is_refskip or self.filter_record(pileupread.alignment,
                                                                                        cell_barcode, umi,
                                                                                        self.min_mapq,
                                                                                        self.max_gene_length):
                        continue

                    if self.filtered_cell_barcodes:
                        if cell_barcode not in self.filtered_cell_barcodes:
                            continue
                    if cell_barcode not in pos_data:
                        pos_data[cell_barcode] = {}
                    if umi not in pos_data[cell_barcode]:
                        pos_data[cell_barcode][umi] = {'mut_reads': 0, 'no_mut_reads': 0}
                    if self.is_mutated_read(pileupread):
                        pos_data[cell_barcode][umi]['mut_reads'] += 1
                    else:
                        pos_data[cell_barcode][umi]['no_mut_reads'] += 1
                valid_cells_num, valid_cells = self.valid_pos(pos_data)
                if valid_cells_num:
                    self.write_pos_data(out_fh, chrom, pileupcolumn.reference_pos + 1, valid_cells_num, valid_cells)
            self.write_pos_data(out_fh)  # last iteration
        print "%s End chromomosome: %s:%s%s " % (datetime.datetime.now(), chrom, start, str_end)

    def write_pos_data(self, out_fh, chrom=None, pos=None, valid_cells_num=None, valid_cells=None):
        if pos:
            cells = ','.join(valid_cells)
            self.valid_positions.append('\t'.join([str(chrom), str(pos), str(valid_cells_num), cells, '\n']))
        if pos and len(self.valid_positions) > 100:
            out_fh.writelines(self.valid_positions)
            self.valid_positions = []
        if not pos:  # last iteration
            out_fh.writelines(self.valid_positions)
            self.valid_positions = []

    def valid_pos(self, pos_data):
        """
        Requiremnts:
        At least 10 valid cells

        valid cell:
        - At least 5 umis
        - No read with mutation
        - There exist at least one umi with at least 2 reads.

        Args:
            pos_data:   (dict): cells:umis:num_mutated_reads
                                          :num_not_mutated_reads

        Returns:
            number of valid cells and list of their cell barcodes
        """
        valid_cells_num = 0
        valid_cells = []
        cells_num = len(pos_data.keys())
        if cells_num < self.min_cells or cells_num > self.max_cells:
            return (False, False)
        for cell in pos_data:
            invalid_cell = False
            umis_num = len(pos_data[cell].keys())
            if (umis_num < self.min_mutated_umis and umis_num < self.min_non_mutated_umis) or (
                    umis_num > self.max_mutated_umis and umis_num > self.max_non_mutated_umis):
                if not self.enable_cells_with_invalid_umis_num:
                    return (False, False)
                else:
                    continue
            valid_mutated_umis_num = 0
            valid_non_mutated_umis_num = 0
            for umi in pos_data[cell]:
                mut_reads_num = pos_data[cell][umi]['mut_reads']
                no_mut_reads_num = pos_data[cell][umi]['no_mut_reads']
                if mut_reads_num and no_mut_reads_num:
                    if not self.enable_umis_with_invalid_reads_num:
                        invalid_cell = True
                        if not self.enable_cells_with_invalid_umis_num:
                            return (False, False)
                        break  # enable invalid cells, but not invalid umis - break from the cell.
                    continue  # enable invalid umi, continue to the next umi
                if mut_reads_num:
                    if mut_reads_num < self.min_reads_per_mutated_umi or mut_reads_num > self.max_reads_per_mutated_umi:
                        if not self.enable_umis_with_invalid_reads_num:
                            invalid_cell = True
                            if not self.enable_cells_with_invalid_umis_num:
                                return (False, False)
                            break  # enable invalid cells, but not invalid umis - break from the cell.
                        continue  # enable invalid umi, continue to the next umi
                    valid_mutated_umis_num += 1
                elif no_mut_reads_num:
                    if no_mut_reads_num < self.min_reads_per_non_mutated_umi or no_mut_reads_num > self.max_reads_per_non_mutated_umi:
                        if not self.enable_umis_with_invalid_reads_num:
                            invalid_cell = True
                            if not self.enable_cells_with_invalid_umis_num:
                                return (False, False)
                            break  # enable invalid cells, but not invalid umis - break from the cell.
                        continue  # enable invalid umi, continue to the next umi
                    valid_non_mutated_umis_num += 1
            if not invalid_cell:
                if (
                        valid_mutated_umis_num >= self.min_mutated_umis and valid_mutated_umis_num <= self.max_mutated_umis) or (
                        valid_non_mutated_umis_num >= self.min_non_mutated_umis and valid_non_mutated_umis_num <= self.max_non_mutated_umis):
                    valid_cells_num += 1
                    valid_cells.append(cell)
                else:
                    if not self.enable_cells_with_invalid_umis_num:
                        return (False, False)
        # print 'num valid cells %s' %valid_cells_num
        if valid_cells_num < self.min_cells or valid_cells_num > self.max_cells:
            return (False, False)
        else:
            return (valid_cells_num, valid_cells)

    def filter_record(self, record, cell_barcode, umi, min_mapq, max_gene_length):
        """
        Args:
            record              (pysam.AlignedSegment object): one read
            min_mapq            (int): Minimum quality of the read mapping
            max_gene_length     (int): Maximum length of the gene. Reads that will be mapped to longer bases will be discarded

        Returns:
            filtered            (bool): True if filtered out, else False
        """

        if re.findall(r"([DSHI]+)", record.cigarstring):  # filter reads with mutation/insertion/softclipped/hardclipped
            return True
        if record.mapq < min_mapq:  # Mapq 10 and above is uniquely mapped
            return True
        if 'N' in umi:
            return True
        if 'N' in cell_barcode:
            return True
        if record.get_tag('XF').startswith('__'):  # filter out reads that didn't mapped to genes
            return True
        if record.reference_length > max_gene_length:
            return True
        else:
            return False



#!/usr/bin/env python

__author__ = 'Refael Kohen'


import argparse
import logging

from singleCellBamQueries.query import Query
from argparse import RawTextHelpFormatter, ArgumentDefaultsHelpFormatter
logger = logging.getLogger(__name__)

class myArgparserFormater(RawTextHelpFormatter, ArgumentDefaultsHelpFormatter):
    """
    RawTextHelpFormatter: can break lines in the help text, but don't print default values
    ArgumentDefaultsHelpFormatter: print default values, but don't break lines in the help text
    """
    pass

def parse_args():
    help_txt = """
    Queries on bam file of single cell per genome position.

    The bam file must be sorted and indexed with the commands:

    samtools sort filename.bam > filename.sorted.bam
    samtools index filename.sorted.bam
    
    By default the umi and barcode cells are comptible to bam files from cellranger. 
    For other formats you need to change the parameters of tags and cell barcodes.  
    
    For now, the script filter out the reads with deletions/insersions/soft and hard clipped.
    In the next version it will be parameter to user.
    """
    # parser = argparse.ArgumentParser(description=help_txt, formatter_class=RawTextHelpFormatter)
    # parser = argparse.ArgumentParser(description=help_txt, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser = argparse.ArgumentParser(description=help_txt, formatter_class=myArgparserFormater)

    parser.add_argument('--input-file', help='Full path to input .bam or .sam file', required=True)
    parser.add_argument('--output-file', help='Full path to output file name', required=True)
    parser.add_argument('--coordinates',
                        help='Coordinates of the genome. The default is all genome. For example: chr1:1000000-2000000, or for all chromosom: chr1',
                        default='all', required=False)
    parser.add_argument('--filtered-cell-barcodes-file',
                        help='Text file with list of cell barcodes. Counts only these cells (optional)', default=None,
                        required=False)
    parser.add_argument('--min-mapq', help='Minimum quality of the read mapping', type=int, default=10, required=False)
    parser.add_argument('--max-gene-length',
                        help='Maximum length of the gene. Reads that will be mapped to longer bases will be discarded',
                        type=int, default=100000, required=False)
    parser.add_argument('--threads',
                        help='number of threads. You can run the chromosome itself in several threads. You can use this prameter only if you specify the start and end coordinates explicitely in the format: chr1:0-14000000 or if the bam file contains header lines with the lengths of the chromosomes, you can check it with the commands: samtools view -h filename.bam',
                        type=int, default=1, required=False)
    parser.add_argument('--min-cells',
                        help='mininum cells in genome position that contains the number of umis and reads according to the other parameters',
                        type=int, default=1, required=False)
    parser.add_argument('--max-cells',
                        help='maximum cells in genome position that contains the number of umis and reads according to the other parameters',
                        type=int, default=1000000000, required=False)
    parser.add_argument('--min-mutated-umis',
                        help='mininum umis per cell that all reads contain mutation in the position',
                        type=int, default=1, required=False)
    parser.add_argument('--max-mutated-umis',
                        help='maximum umis per cell that all reads contain mutation in the position',
                        type=int, default=1000000000, required=False)
    parser.add_argument('--min-reads-per-non-mutated-umi',
                        help='mininum reads in at least one of umis in the cell in genome position',
                        type=int, default=1, required=False)
    parser.add_argument('--max-reads-per-non-mutated-umi',
                        help='maximum reads in at least one of umis in the cell in genome position',
                        type=int, default=1000000000, required=False)
    parser.add_argument('--min-non-mutated-umis',
                        help='mininum umis per cell that all reads not contain mutation in the genome position',
                        type=int, default=1, required=False)
    parser.add_argument('--max-non-mutated-umis',
                        help='maximum umis per cell that all reads not contain mutation in the genome position',
                        type=int, default=1000000000, required=False)
    parser.add_argument('--min-reads-per-mutated-umi',
                        help='mininum reads in at least one of umis in the cell in genome position',
                        type=int, default=1, required=False)
    parser.add_argument('--max-reads-per-mutated-umi',
                        help='maximum reads in at least one of umis in the cell in genome position',
                        type=int, default=1000000000, required=False)
    parser.add_argument('--enable-cells-with-invalid-umis-num',
                        help='enable positions that contain cell/s with not valid umis number (according to ther range in the other parameters)',
                        action='store_true', default=False, required=False)
    parser.add_argument('--enable-umis-with-invalid-reads-num',
                        help='enable positions that contain umi/s with not valid reads number (according to ther range in the other parameters)',
                        action='store_true', default=False, required=False)
    parser.add_argument('--tag-of-umi', help='the tag of umi in bam file', type=str, default='UR', required=False)
    parser.add_argument('--tag-of-cell-barcode', help='the tag of umi in bam file', type=str, default='CR', required=False)
    parser.add_argument('--umi-start', help='location in tag where the umi start (0-based)', type=int, default=0, required=False)
    parser.add_argument('--umi-length', help='length of umi', type=int, default=10, required=False)
    parser.add_argument('--cell-barcode-start', help='location in tag where the cell barcode start (0-based)', type=int, default=0, required=False)
    parser.add_argument('--cell-barcode-length', help='length of cell barcode', type=int, default=16, required=False)
    parser.add_argument('--log-file', help='Log File', default=None, required=False)
    return parser.parse_args()


if __name__ == "__main__":
    logging_args = {
        "level": logging.DEBUG,
        "filemode": 'w',
        "format": '%(asctime)s %(message)s',
        "datefmt": '%m/%d/%Y %H:%M:%S'
    }
    args = parse_args()

    # set up log file
    if args.log_file is not None:
        logging_args["filename"] = args.logFile
    logging.basicConfig(**logging_args)

    logging.info('Program started')

    Query(args.input_file, args.output_file, args.coordinates, args.filtered_cell_barcodes_file,
          args.min_mapq, args.max_gene_length, args.threads, args.min_cells, args.max_cells,
          args.min_mutated_umis, args.max_mutated_umis, args.min_reads_per_mutated_umi,
          args.max_reads_per_mutated_umi, args.min_non_mutated_umis, args.max_non_mutated_umis,
          args.min_reads_per_non_mutated_umi, args.max_reads_per_non_mutated_umi,
          args.enable_cells_with_invalid_umis_num,
          args.enable_umis_with_invalid_reads_num, args.tag_of_umi, args.tag_of_cell_barcode, args.umi_start,
          args.umi_length, args.cell_barcode_start, args.cell_barcode_length).run()

    logging.info('Program finished')

"""
#Run on sura:
cd /data/users/pmrefael/workspace/single-cell-bam-queries/singleCellBamQueries/
python /data/users/pmrefael/workspace/single-cell-bam-queries/singleCellBamQueries/single-cell-bam-queries.py --input-file /data/users/pmrefael/workspace/rvcu/tests-statistics/input-data-bigfiles/TTAGGACCAGCAGTTT.sort.bam --output-file /data/users/pmrefael/workspace/rvcu/tests-statistics/output-data/non-mutated_bases --filtered-cell-barcodes-file /data/users/pmrefael/workspace/rvcu/tests-statistics/input-data-bigfiles/pp1-barcodes.tsv --min-mapq 10 --max-gene-length 100000 --threads 1 --tag-of-umi RX --tag-of-cell-barcode RX --umi-start 16 --umi-length 10 --cell-barcode-start 0 --cell-barcode-length 16

#bam must be sorted and indexed.
samtools sort ...bam > ...sorted.bam
samtools index ...sorted.bam

#Filter barcodes files:
/home/labs/gillev/Collaboration/180417_NB501465_0284_AHNYLHBGX5_10X/cell_ranger_out/micePP1_force1500/outs/filtered_gene_bc_matrices/mm10/barcodes.tsv
/home/labs/gillev/Collaboration/180417_NB501465_0284_AHNYLHBGX5_10X/cell_ranger_out/micePP2_force1500/outs/filtered_gene_bc_matrices/mm10/barcodes.tsv

"""
#TODO:
#Support also reads with deletions/insertions/clipped
#Filter reads: user will can select the filtering.








import logging
import multiprocessing
from math import ceil, floor

from positions_query import PositionsQuery

logger = logging.getLogger(__name__)


class Query(object):
    """
    Read bam input file sorted by coordinates. Read the records of one chromosom.

    Args:
        bam_input                               (str): path to input file (bam format sorted by coordinates). Must to be bam.bai file in the same folder
        output_file                             (str): path to output file
        min_mapq                                (int): Minimum quality of the read mapping
        max_gene_length                         (int): Maximum length of the gene. Reads that will be mapped to longer bases will be discarded
        threads                                 (int): number of threads in each chromosome. The program run each chromose in parallel. You can run the chromosome itself in several threads. You can use this prameter only if you specify the start and end coordinates explicitely in the format: chr1:0-14000000 or if the bam file contains header lines with the lengths of the chromosomes, you can check it with the commands: samtools view -h filename.bam
        min_cells                               (int): mininum cells in genome position that contains the number of umis and reads according to the other parameters
        max_cells                               (int): maximum cells in genome position that contains the number of umis and reads according to the other parameters
        min_mutated_umis                        (int): mininum umis per cell that all reads contain mutation in the position
        max_mutated_umis                        (int): maximum umis per cell that all reads contain mutation in the position
        min_reads_per_mutated_umi               (int): mininum reads in at least one of umis in the cell in genome position
        max_reads_per_mutated_umi               (int): maximum reads in at least one of umis in the cell in genome position
        min_non_mutated_umis                    (int): mininum umis per cell that all reads not contain mutation in the genome position
        max_non_mutated_umis                    (int): maximum umis per cell that all reads not contain mutation in the genome position
        min_reads_per_non_mutated_umi           (int): mininum reads in at least one of umis in the cell in genome position
        max_reads_per_non_mutated_umi           (int): maximum reads in at least one of umis in the cell in genome position
        enable_cells_with_invalid_umis_num    (int): enable positions that contain cell/s with not valid umis number (according to ther range in the other parameters)
        enable_umis_with_invalid_reads_num    (int): enable positions that contain umi/s with not valid reads number (according to ther range in the other parameters)
    """

    def __init__(self, input_file, output_file, coordinates, filtered_cell_barcodes_file, min_mapq, max_gene_length,
                 threads, min_cells, max_cells, min_mutated_umis, max_mutated_umis, min_reads_per_mutated_umi,
                 max_reads_per_mutated_umi, min_non_mutated_umis, max_non_mutated_umis, min_reads_per_non_mutated_umi,
                 max_reads_per_non_mutated_umi, enable_cells_with_invalid_umis_num,
                 enable_umis_with_invalid_reads_num, tag_of_umi, tag_of_cell_barcode, umi_start, umi_length,
                 cell_barcode_start, cell_barcode_length):
        self.input_file = input_file
        self.output_file = output_file
        self.coordinates = self.parse_coordinates(coordinates)
        self.min_mapq = min_mapq
        self.max_gene_length = max_gene_length
        self.threads = threads
        self.min_cells = min_cells
        self.max_cells = max_cells
        self.min_mutated_umis = min_mutated_umis
        self.max_mutated_umis = max_mutated_umis
        self.min_reads_per_mutated_umi = min_reads_per_mutated_umi
        self.max_reads_per_mutated_umi = max_reads_per_mutated_umi
        self.min_non_mutated_umis = min_non_mutated_umis
        self.max_non_mutated_umis = max_non_mutated_umis
        self.min_reads_per_non_mutated_umi = min_reads_per_non_mutated_umi
        self.max_reads_per_non_mutated_umi = max_reads_per_non_mutated_umi
        self.filtered_cell_barcodes_file = filtered_cell_barcodes_file
        self.enable_cells_with_invalid_umis_num = enable_cells_with_invalid_umis_num
        self.enable_umis_with_invalid_reads_num = enable_umis_with_invalid_reads_num
        self.tag_of_umi = tag_of_umi
        self.tag_of_cell_barcode = tag_of_cell_barcode
        self.umi_start = umi_start
        self.umi_length = umi_length
        self.cell_barcode_start = cell_barcode_start
        self.cell_barcode_length = cell_barcode_length

        if filtered_cell_barcodes_file:
            self.filtered_cell_barcodes = self.find_valid_cells()

    def find_valid_cells(self):
        filtered_cell_barcodes = {}
        with open(self.filtered_cell_barcodes_file, 'r') as filtered_cell_barcodes_fh:
            for line in filtered_cell_barcodes_fh:
                line = line.strip()
                if line.endswith("-1"):
                    line = line[:-2]  # Remove -1 characters in end of line
                filtered_cell_barcodes[line] = None
        return filtered_cell_barcodes

    def parse_coordinates(self, coordinates):
        if coordinates == 'all':
            return [coordinates]
        import re
        # coordinates = 'chr1'
        # coordinates = 'chr1:223-444'
        chrom = re.compile('(\S+)')
        coor = re.compile('(\S+):(\d+)-(\d+)')
        if coor.match(coordinates):
            return coor.match(coordinates).groups()
        elif chrom.match(coordinates):
            return [chrom.match(coordinates).groups()[0]]
        else:
            raise IOError("The coordinates are not valid")

    def run(self):
        chrom_list, starts, ends = [], [], []
        # if need to get the lengths of the chromosomes from the bam file
        if self.coordinates == ['all'] or (self.threads > 1 and len(self.coordinates) != 3):
            chrom_list, starts, ends = PositionsQuery.get_chrom_list(self.input_file)
        if self.coordinates != ['all'] and len(self.coordinates) == 1:
            if self.threads > 1:  # get the chrom length
                for c, l in zip(chrom_list, ends):  # chrom_list already defined in previous "if"
                    if self.coordinates[0] == c:
                        ends = [l]
                        break
            else:
                ends = [None]

            chrom_list, starts = [self.coordinates[0]], [0]
        if len(self.coordinates) == 3:
            chrom_list, starts, ends = [self.coordinates[0]], [int(self.coordinates[1])], [int(self.coordinates[2])]

        if self.threads > len(chrom_list):
            segment_num_per_chrom = int(floor(self.threads / len(chrom_list)))
            cutted_chrom_list = []
            cutted_starts = []
            cutted_ends = []
            for chrom, start, end in zip(chrom_list, starts, ends):
                segment_size = int(ceil(float(end - start + 1) / segment_num_per_chrom))
                s = start
                for i in xrange(segment_num_per_chrom):
                    cutted_chrom_list.append(chrom)
                    cutted_starts.append(s)
                    cutted_ends.append(min(s + segment_size, end))
                    s += segment_size + 1
            chrom_list = cutted_chrom_list
            starts = cutted_starts
            ends = cutted_ends

        logger.info('The program run with %s threads in parallel on the segments:' % self.threads)
        for chrom, start, end in zip(chrom_list, starts, ends):
            logger.info('%s:%s-%s' % (chrom, start, end))

        pool = multiprocessing.Pool(processes=self.threads)
        for chrom, start, end in zip(chrom_list, starts, ends):
            bam_reader = PositionsQuery(self.input_file, self.output_file, self.filtered_cell_barcodes,
                                        self.min_mapq,
                                        self.max_gene_length, self.min_cells, self.max_cells, self.min_mutated_umis,
                                        self.max_mutated_umis, self.min_reads_per_mutated_umi,
                                        self.max_reads_per_mutated_umi, self.min_non_mutated_umis,
                                        self.max_non_mutated_umis,
                                        self.min_reads_per_non_mutated_umi, self.max_reads_per_non_mutated_umi,
                                        self.enable_cells_with_invalid_umis_num,
                                        self.enable_umis_with_invalid_reads_num,
                                        self.tag_of_umi, self.tag_of_cell_barcode, self.umi_start, self.umi_length,
                                        self.cell_barcode_start, self.cell_barcode_length)
            pool.apply_async(bam_reader.find_base_coverage, args=(chrom, start, end))
        pool.close()
        pool.join()



from unittest import TestCase

from singleCellBamQueries.positions_query import PositionsQuery


class TestPositionsQuery(TestCase):
    def setUp(self):
        pos_q = PositionsQuery(*(17 * [None]))

    def test_valid_pos(self):
        pos_q = PositionsQuery(None, None, None, None, None,
                               min_cells=1, max_cells=2,
                               min_mutated_umis=1, max_mutated_umis=2,
                               min_reads_per_mutated_umi=2, max_reads_per_mutated_umi=3,
                               min_non_mutated_umis=1, max_non_mutated_umis=2,
                               min_reads_per_non_mutated_umi=2, max_reads_per_non_mutated_umi=3,
                               enable_cells_with_invalid_umis_num=False,
                               enable_umis_with_invalid_reads_num=False,
                               tag_of_umi=None, tag_of_cell_barcode=None,umi_start=None, umi_length=None,
                               cell_barcode_start=None, cell_barcode_length=None)

        # valid case:
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))

        # enable_cells_with_invalid_umis_num = False
        # enable_umis_with_invalid_reads_num = False
        # UMI with mutated and not mutated reads
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 2}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))
        # one umi with reads above max in mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 4, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))
        # one umi with reads under min in mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 1, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))
        # one umi with reads above max in not mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 4}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))
        # one umi with reads under min in not mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 1}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))
        # valid one umi
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # no umi
        pos_data = {'AAA': {},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))
        # whole cell invalid
        pos_data = {'AAA': {'GG': {'mut_reads': 1, 'no_mut_reads': 0}, 'TT': {'mut_reads': 1, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))
        # whole cell invalid
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 1}, 'TT': {'mut_reads': 0, 'no_mut_reads': 1}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))

        pos_q = PositionsQuery(None, None, None, None, None,
                               min_cells=1, max_cells=2,
                               min_mutated_umis=1, max_mutated_umis=2,
                               min_reads_per_mutated_umi=2, max_reads_per_mutated_umi=3,
                               min_non_mutated_umis=1, max_non_mutated_umis=2,
                               min_reads_per_non_mutated_umi=2, max_reads_per_non_mutated_umi=3,
                               enable_cells_with_invalid_umis_num=False,
                               enable_umis_with_invalid_reads_num=True,
                               tag_of_umi=None, tag_of_cell_barcode=None,umi_start=None, umi_length=None,
                               cell_barcode_start=None, cell_barcode_length=None)

        # enable_cells_with_invalid_umis_num = False
        # enable_umis_with_invalid_reads_num = True
        # UMI with mutated and not mutated reads
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 2}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # one umi with reads above max in mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 4, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # one umi with reads under min in mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 1, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # one umi with reads above max in not mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 4}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # one umi with reads under min in not mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 1}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # no umi
        pos_data = {'AAA': {},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))
        # whole cell invalid
        pos_data = {'AAA': {'GG': {'mut_reads': 1, 'no_mut_reads': 0}, 'TT': {'mut_reads': 1, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))
        # whole cell invalid
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 1}, 'TT': {'mut_reads': 0, 'no_mut_reads': 1}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (False, False))

        pos_q = PositionsQuery(None, None, None, None, None,
                               min_cells=1, max_cells=2,
                               min_mutated_umis=1, max_mutated_umis=2,
                               min_reads_per_mutated_umi=2, max_reads_per_mutated_umi=3,
                               min_non_mutated_umis=1, max_non_mutated_umis=2,
                               min_reads_per_non_mutated_umi=2, max_reads_per_non_mutated_umi=3,
                               enable_cells_with_invalid_umis_num=True,
                               enable_umis_with_invalid_reads_num=False,
                               tag_of_umi=None, tag_of_cell_barcode=None,umi_start=None, umi_length=None,
                               cell_barcode_start=None, cell_barcode_length=None)

        # enable_cells_with_invalid_umis_num = True
        # enable_umis_with_invalid_reads_num = False
        # UMI with mutated and not mutated reads
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 2}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['CCC']))
        # one umi with reads above max in mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 4, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['CCC']))
        # one umi with reads under min in mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 1, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['CCC']))
        # one umi with reads above max in not mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 4}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['AAA']))
        # one umi with reads under min in not mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 1}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['AAA']))
        # no umi
        pos_data = {'AAA': {},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['CCC']))
        # whole cell invalid
        pos_data = {'AAA': {'GG': {'mut_reads': 1, 'no_mut_reads': 0}, 'TT': {'mut_reads': 1, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['CCC']))
        # whole cell invalid
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 1}, 'TT': {'mut_reads': 0, 'no_mut_reads': 1}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['AAA']))

        pos_q = PositionsQuery(None, None, None, None, None,
                               min_cells=1, max_cells=2,
                               min_mutated_umis=1, max_mutated_umis=2,
                               min_reads_per_mutated_umi=2, max_reads_per_mutated_umi=3,
                               min_non_mutated_umis=1, max_non_mutated_umis=2,
                               min_reads_per_non_mutated_umi=2, max_reads_per_non_mutated_umi=3,
                               enable_cells_with_invalid_umis_num=True,
                               enable_umis_with_invalid_reads_num=True,
                               tag_of_umi=None, tag_of_cell_barcode=None,umi_start=None, umi_length=None,
                               cell_barcode_start=None, cell_barcode_length=None)

        # enable_cells_with_invalid_umis_num = True
        # enable_umis_with_invalid_reads_num = True
        # UMI with mutated and not mutated reads
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 2}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # one umi with reads above max in mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 4, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # one umi with reads under min in mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 1, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # one umi with reads above max in not mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 4}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # one umi with reads under min in not mut umi
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 1}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (2, ['AAA', 'CCC']))
        # no umi
        pos_data = {'AAA': {},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['CCC']))
        # whole cell invalid
        pos_data = {'AAA': {'GG': {'mut_reads': 1, 'no_mut_reads': 0}, 'TT': {'mut_reads': 1, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 2}, 'TT': {'mut_reads': 0, 'no_mut_reads': 2}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['CCC']))
        # whole cell invalid
        pos_data = {'AAA': {'GG': {'mut_reads': 2, 'no_mut_reads': 0}, 'TT': {'mut_reads': 2, 'no_mut_reads': 0}},
                    'CCC': {'GG': {'mut_reads': 0, 'no_mut_reads': 1}, 'TT': {'mut_reads': 0, 'no_mut_reads': 1}}}
        self.assertEqual(pos_q.valid_pos(pos_data), (1, ['AAA']))



from setuptools import setup, find_packages

with open('VERSION.txt', 'r') as version_file:
    version = version_file.read().strip()

requires = ['']

setup(
    name='bbcu.singleCellBamQueries',
    version=version,
    author='Refael Kohen',
    author_email='refael.kohen@weizmann.ac.il',
    packages=find_packages(),
    scripts=[
        'scripts/single-cell-bam-queries.py',
    ],
    description='Queries on bam file of single cell per genome position',
    long_description=open('README.txt').read(),
    long_description_content_type='text/markdown',
    install_requires=requires,
    tests_require=requires + ['nose'],
    include_package_data=True,
    test_suite='nose.collector',
)
